      subroutine dirsol(
     .  rhsid,unkno,ntotv,nsist,kfact,iprob,outso)
c****************************************************************************
c
c**** This routine drives the library to solve a linear system of equations
c**** by direct methods using the LDU decomposition. This decomposition is
c**** done only when KFACT = 1. Otherwise, only a backward and a forward
c**** substituion is performed.       
c
c****************************************************************************
      implicit none
      include   'MatMan.h'
      integer    ntotv,nsist,iprob,multi,ifail,kfact,
     .           iflag
      real*8     rhsid(ntotv,nsist), unkno(ntotv,nsist)
      integer    p_mat,iwoso,imasp,p_col,p_row,p_lpn,       ! pointer
     .           p_lpo,p_rof,p_apc,p_rhs
      logical*1  outso
      integer    lbyta,lbyts,svomt
      data       lbyta,svomt/0,0/
      save       lbyta,iwoso,svomt
c
c*** Allocate a work space for the redefinition of RHSID and UNKNO
c
      lbyts=8*ntotv*nsist
      if (lbyta.eq.0) then
        call sadmem(0,iwoso,lbyts,svomt)
        lbyta=lbyts
      else
        if (lbyts.gt.lbyta) then
          svome(iprob)=max(svome(iprob),svomt)        
          call sadmem(2,iwoso,lbyta,svomt)
          call sadmem(0,iwoso,lbyts,svomt)
          lbyta=lbyts
        end if
      end if
      call spoias(%val(ialpn+(iprob-1)*lengp),p_lpn)
c
c*** Compute A_p x_p, where x_p are the prescribed values of the unknown
c
      iflag=1
      call spoias(%val(iapma+(iprob-1)*lengp),p_mat)
      if(p_mat.eq.0) iflag=0
      call spoias(%val(irhsx+(iprob-1)*lengp),p_rhs)
      if((p_mat.ne.0).and.(kfact.le.1)) then
        call spoias(%val(iapii+(iprob-1)*lengp),p_rof)
        call spoias(%val(iapij+(iprob-1)*lengp),p_apc)
        call modrhs(
     .    %val(p_lpn),%val(p_rof),%val(p_apc),
     .    unkno,%val(p_rhs),%val(p_mat),ntotv,
     .    nzecp(iprob),nsist,nequa(iprob),iflag,
     .    zeroc)
      end if
c
c*** Redefine RHSID according to the renumbering strategy
c
      if(iflag.eq.0) p_rhs=iwoso                            ! not used 
      call redefr(
     .  %val(iwoso),%val(p_lpn),%val(p_rhs),rhsid,ntotv,
     .  nequa(iprob),nsist,iflag)
c
c*** Solve the final algebraic system using a band storage
c
      call spoias(%val(iamat+(iprob-1)*lengp),p_mat)
      if(kmsip(iprob).eq.0) then
        if(ksymm(iprob).eq.0) then
          if(kites(iprob).eq.11) then
            call asbddd(nequa(iprob),nband(iprob),%val(p_mat),
     .        %val(p_mat),nsist,%val(iwoso),%val(iwoso),
     .        kfact,lusol)
          else if(kites(iprob).eq.12) then
            call srunen(
     .        'DIRSOL: PIVOTING NOT YET IMPLEMENTED',lusol)
          end if            
        else if(ksymm(iprob).eq.1) then
          if(kites(iprob).eq.11) then
            call aubidd(nequa(iprob),nband(iprob),%val(p_mat),
     .        %val(p_mat),nsist,%val(iwoso),%val(iwoso),
     .        kfact,lusol)
          else if(kites(iprob).eq.12) then
            call srunen(
     .        'DIRSOL: PIVOTING NOT YET IMPLEMENTED',lusol)
          end if            
        end if
c
c*** Solve the final algebraic system using a profile storage
c
      else if (kmsip(iprob).eq.1) then
        call spoias(%val(ialpo+(iprob-1)*lengp),p_lpo)
        if(kites(iprob).eq.11) then
          call skylin(nequa(iprob),%val(p_mat),%val(iwoso),
     .      %val(p_lpo),nsist,ksymm(iprob),kfact,
     .      lusol)
        else if(kites(iprob).eq.12) then
          call srunen('DIRSOL: PIVOTING NOT YET IMPLEMENTED',lusol)
        end if          
c
c*** Solve the final algebraic system using a sparse storage
c
      else if (kmsip(iprob).eq.2) then
        if(kites(iprob).eq.11) then
          call srunen(
     .      'DIRSOL: ONLY THE PIVOTING VERSION AVAILABLE',lusol)
        else if(kites(iprob).eq.12) then
          call canonz(nequa(iprob),nband(iprob),
     .      %val(p_mat),
     .      nonze(iprob),ksymm(iprob))
          multi=nmult(iprob)
          ifail=5
          do while (ifail.eq.5)
            call sadmem(0,imasp,multi*nonze(iprob)*8    ,svomt)         
            call sadmem(0,p_col,multi*nonze(iprob)*lengp,svomt)
            call spoias(p_col,%val(incol+(iprob-1)*lengp))
            call sadmem(0,p_row,multi*nonze(iprob)*lengp,svomt)
            call spoias(p_row,%val(inrow+(iprob-1)*lengp))
            call batosp(nequa(iprob),nband(iprob),
     .        %val(p_mat),%val(imasp),
     .        %val(p_col),%val(p_row),
     .        nonze(iprob),multi,ksymm(iprob),%val(iwoso))
            call sparse(
     .        nequa(iprob),nonze(iprob),multi,%val(imasp),
     .        %val(p_col),%val(p_row),%val(iwoso),
     .        nsist,ifail,kfact,lusol,svomt,svome(iprob),
     .        lengp)
            svome(iprob)=max(svome(iprob),svomt)        
            call sadmem(2,imasp,multi*nonze(iprob)*8    ,svomt) ! Deal. temp.
            call sadmem(2,p_col,multi*nonze(iprob)*lengp,svomt) ! memory
            call sadmem(2,p_row,multi*nonze(iprob)*lengp,svomt)
            if (ifail.ne.0)
     .        call redefr(%val(iwoso),%val(p_lpn),%val(p_rhs),
     .        rhsid,ntotv,nequa(iprob),nsist,iflag)
            multi=multi+1
          end do
          nmult(iprob)=multi-1
        end if
      end if
c
c*** Redefine UNKNO according to the renumbering strategy
c
      call redunk(%val(iwoso),%val(p_lpn),unkno,
     .            ntotv,nequa(iprob),nsist)
c
c*** Write temporary memory requirements 
c
      if(outso) write(lusol,110) svome(iprob)
c
c*** Formats
c
  110 format(11x,'*** VOLATILE MEMORY (BYTES): ',i12)

      end
