      subroutine skybak(gstdi,gstlo,gstup,lpont,neqns,refor,lures)
c***********************************************************************
c
c**** This routine performs the back-solution of symmetric equations
c     stored in profile form .
c     coefficient matrix must be decomposed into its triangular
c     factors using skylin before using skybak.
c
c.... input parameters
c
c         gstdi(neqns)       -reciprocal of diagonal of triangular 
c                               factor
c         gstlo(lpont(neqns) -lower triangular factor of matrix
c         gstup(lpont(neqns) -upper triangular factor of matrix
c                               (gstup and gstlo have same calling 
c                               address for symmetric matrices)
c         refor(neqns)       -right hand side vector in equations
c         lpont(neqns)       -pointer array to bottom of columns 
c                               of gstlo and gstup
c         neqns              -number of equations to be solved
c
c.... output parameter
c
c         refor(neqns)       -solution of equations
c
c***********************************************************************
      implicit real*8(a-h,o-z)
      integer  lures
      real*8   gstlo(*), gstup(*), gstdi(*), refor(*)
      integer  lpont(*)
      zeros=0.0d0
c
c*** Find the first nonzero entry in the right hand side
c
      do 100 keqns=1,neqns
        ieqns=keqns
        if(refor(ieqns).ne.zeros) go to 200
  100 continue
      if(neqns.gt.0) write(lures,2000)
      return
  200 if(ieqns.lt.neqns) then
c
c*** Reduce the right hand side
c
        do 300 jeqns=ieqns+1,neqns
          jrows=lpont(jeqns-1)
          jheig=lpont(jeqns)-jrows
          if(jheig.gt.0) then
            refor(jeqns)=refor(jeqns)-
     .        svecdo(gstlo(jrows+1),refor(jeqns-jheig),jheig)
          endif
  300   continue
      endif
c
c*** Multiply by inverse of diagonal elements
c
      do 400 jeqns=ieqns,neqns
        refor(jeqns)=refor(jeqns)*gstdi(jeqns)
  400 continue
c
c*** Backsubstitution
c
      if(neqns.gt.1) then
        do 500 jeqns=neqns,2,-1
          jrows=lpont(jeqns-1)
          jheig=lpont(jeqns)-jrows
          if(jheig.gt.0) then
            kterm=jeqns-jheig
            do 600 iterm=1,jheig
              refor(kterm)=refor(kterm)-gstup(jrows+iterm)*refor(jeqns)
              kterm=kterm+1
  600       continue
          endif
  500   continue
      endif
      return
 2000 format(11x,'*** WARNING: ZERO RIGHT-HAND-SIDE VECTOR')
      end
      
