      subroutine skytri(gstdi,gstlo,gstup,isymm,lpont,neqns,lures)
c***********************************************************************
c
c*** This routine performs the triangular decomposition 
c    of a matrix stored in profile form
c
c....input parameters
c
c       gstlo(lpont(neqns)) - lower triangular part of matrix
c       gstup(lpont(neqns)) - upper part of triangular matrix
c       gstdi(neqns)        - diagonals of triangular matrix
c       lpont(neqns)        - pointers to bottom of colums of 
c                             gstlo and gstup arrays
c       neqns               - number of equations to be solved
c       isymm               - =0 equations are unsymmetric
c                             =1 equations are symmetric and calling
c                                 address of gstlo may be same 
c                                 as that for gstup (i.e., gstup 
c                                 and gstlo share same memory)
c
c.... output parameters
c
c       gstlo(lpont(neqns)) - lower triangular factor of matrix
c       gstup(lpont(neqns)) - upper triangular factor of matrix
c       gstdi(neqns)        - inverse of diagonal matrix in triangular 
c                              factor
c
c.... local  parameters
c
c       tolle               - tollerance used to check for null pivot
c                              it should be set to approximate half-word
c                              precision 
c
c***********************************************************************
      implicit real*8(a-h,o-z)
      integer  lpont(neqns)
      real*8   gstlo(*),                                    !gstlo(lpont(neqns))
     .         gstup(*),                                    !gstup(lpont(neqns))
     .         gstdi(neqns)

      nillc=0                                               ! Number of ILL-C
      zeros=0.0d0
      unity=1.0d0
      tolle=0.5d-07
c
c*** If isymm=0 or isymm=1 perform triangular decomposition
c
c*** Loop through the columns to perform the triangular decomposition
c
      kpivo=0
      idiag=1
      do 200 ieqns=1,neqns
        irows=idiag+1
        idiag=lpont(ieqns)
        iheig=idiag-irows
        if(iheig.gt.0) then
          istar=ieqns-iheig
          iends=ieqns-1
c
c*** If diagonal is zero compute a norm for singularity test
c
          if(gstdi(ieqns).eq.zeros) 
     .      call skycek(gstup(irows),iheig,saval)
          do 100 jeqns=istar,iends
            irows=irows+1
            jdiag=lpont(jeqns)
            jheig=min0(jdiag-lpont(jeqns-1),jeqns-istar+1)
            if(jheig.gt.0) then
              irhei=irows-jheig
              jdhei=jdiag-jheig+1
              gstup(irows)=gstup(irows)-
     .          svecdo(gstup(irhei),gstlo(jdhei),jheig) 
              if(isymm.eq.0) 
     .          gstlo(irows)=gstlo(irows)-
     .          svecdo(gstlo(irhei),gstup(jdhei),jheig) 
            end if
  100     continue
        end if
c
c*** Reduce the diagonal
c
        if(iheig.ge.0) then
          pivot=gstdi(ieqns)
          irows=idiag-iheig
          irhei=ieqns-iheig-1
          call skydia(gstlo(irows),gstup(irows),gstdi(irhei),iheig+1,
     .      isymm,gstdi(ieqns))
c
c*** Check for possible errors and print warnings
c
          if(pivot.lt.zeros) kpivo=kpivo+1
          if(abs(gstdi(ieqns)).lt.tolle*abs(pivot))
     .      nillc=nillc+1
          if(abs(gstdi(ieqns)).lt.1.0d-15)
     .      write(lures,2001) ieqns
c
c*** Complete rank test for a zero diagonal case
c
          if(pivot.eq.zeros.and.iheig.gt.0) then
            if(abs(gstdi(ieqns)).lt.tolle*saval) then
              write(lures,2003) ieqns
              stop
            end if
          end if
        end if
c
c*** Store reciprocal of diagonal
c
        if(gstdi(ieqns).ne.zeros) gstdi(ieqns)=unity/gstdi(ieqns)

  200 continue
      if (nillc.ne.0) write(lures,2000) nillc
c
c*** Formats
c
 2000 format(11x,'*** ILL-CONDITIONING.',
     .  ' LOSS OF AT LEAST 7 DIGITS FOR ',I8,' EQS.')
 2001 format(11x,'*** SINGULAR MATRIX.',
     .  ' REDUCED DIAGONAL IS ZERO FOR EQ.',I5)
 2003 format(11x,'*** SINGULAR MATRIX.',
     .  ' RANK FAILURE FOR ZERO UNREDUCED DIAGONAL IN EQ. ',I5)

      end
