      subroutine cgprec(lnods,eamat,lengp,adiag,x_k,bvect,toler,kmaxi,
     .                  r_k_1,p_k,z_k_1,w_k,elpve,elwve,nnode,nnodt,
     .                  nnodw,nelem,nequa,nevaw,ndofn,npoin,npois,
     .                  iffix,lures,outso)
c****************************************************************************
c
c**** This routine solves a linear system of equations by the CG method.
c**** A diagonal preconditioning is used.
c
c****************************************************************************
      implicit none
      integer   eamat                                       ! pointer
      integer   nnode,nnodt,nnodw,nelem,nequa,nevaw,ndofn,npoin,
     .          npois,iequa,kiter,kmaxi,lengp,lures
      real*8    toler,bnor2,rnor2,tole1,beta1,beta2,betak,alfak,
     .          reini,refin,rerel
      integer   lnods(nnodt,nelem), iffix(nequa)
      real*8    adiag(nequa), bvect(nequa), x_k(nequa),
     .          r_k_1(nequa), p_k(nequa),   z_k_1(nequa), 
     .          w_k(nequa),   elpve(nevaw), elwve(nevaw)
      logical*1 outso

      call mabvec(lnods,eamat,lengp,x_k,w_k,elpve,elwve,    ! w_k=A*x_k
     .            nequa,nevaw,nnode,nnodt,nnodw,ndofn,
     .            npoin,npois,nelem, 0)
      do iequa=1,nequa
        if(abs(adiag(iequa)).lt.1.d-15) adiag(iequa)=1.0
        r_k_1(iequa)=bvect(iequa)-w_k(iequa)                ! r_0=b-A*x
      end do
      bnor2=0.0d0
      rnor2=0.0d0
      do iequa=1,nequa
        if(iffix(iequa).ne.1) then
          bnor2=bnor2+bvect(iequa)*bvect(iequa)
          rnor2=rnor2+r_k_1(iequa)*r_k_1(iequa)
        end if
      end do 
      reini=sqrt(rnor2/bnor2)
      kiter=0                                               ! k=0
      tole1=toler*toler*bnor2

      do while ((rnor2.gt.tole1.and.kiter.lt.kmaxi).or.kiter.eq.0)
        kiter=kiter+1                                       ! k=k+1
        do iequa=1,nequa
          z_k_1(iequa)=r_k_1(iequa)/adiag(iequa)            ! Solve M*z_k=r_k
        end do
        if (kiter.eq.1) then
          beta1=0.0d0
          do iequa=1,nequa
            p_k(iequa)=z_k_1(iequa)                         ! p_1=z_0
            beta1=beta1+r_k_1(iequa)*z_k_1(iequa)           ! r_0*z_0
          end do
        else
          beta2=beta1
          beta1=0.0d0
          do iequa=1,nequa
            beta1=beta1+r_k_1(iequa)*z_k_1(iequa)           ! r_k-1*z_k-1
          end do
          betak=beta1/beta2                                 ! B_k
          do iequa=1,nequa
            p_k(iequa)=z_k_1(iequa)+betak*p_k(iequa)        ! p_k=z_k-1+B_k*p_k-1
          end do
        end if
        call mabvec(lnods,eamat,lengp,p_k,w_k,elpve,elwve,  ! w_k=A*p_k
     .              nequa,nevaw,nnode,nnodt,nnodw,ndofn,
     .              npoin,npois,nelem, 0)
        alfak=0.0d0
        do iequa=1,nequa
          alfak=alfak+p_k(iequa)*w_k(iequa)                 ! p_k*w_k
        end do
        alfak=beta1/alfak                                   ! Alfa_k
        rerel=rnor2
        rnor2=0.0
        do iequa=1,nequa
          x_k(iequa)=x_k(iequa)+alfak*p_k(iequa)      ! x_k=x_k-1 + Alfa_k*p_k
          r_k_1(iequa)=r_k_1(iequa)-alfak*w_k(iequa)  ! r_k=r_k-1 + Alfa_k*w_k
          rnor2=rnor2+r_k_1(iequa)*r_k_1(iequa)             ! rnor2=r_k*r_k
        end do
      end do

      refin=sqrt(rnor2/bnor2)
      rerel=sqrt(rnor2/rerel)
      if(outso) write(lures,100) reini,refin,rerel,kiter

  100 format(11x,'*** INITIAL RESIDUAL NORM  : ',e12.5,/,
     .       11x,'*** FINAL RESIDUAL NORM    : ',e12.5,/,
     .       11x,'*** LAST RESIDUAL FACTOR   : ',e12.5,/,
     .       11x,'*** NUMBER OF ITERATIONS   : ',i12)

      end
