      subroutine gmrcor(eamat,lengp,diag,work,work1,x, idim, k0, 
     .                  itmax, toler, yours, iprint, lnods,
     .                  elpve,elwve,nevaw,nnode,nnodt,nnodw,
     .                  ndofn,npoin,npois,nelem,ksymm,lures,
     .                  nwork,iffix)
c*************************************************************************
c
c**** Basic routine of the GMRES iterative solver
c      
c     DATA:
c      
c     idim   : dimension of the problem (the matrix is (idim,idim))
c     k0     : dimension of the krylov subspace
c     itmax  : maximal number of iterations
c     toler  : relative precision needed to stop the process
c     Ax     : external procedure wich computes Ax or b-Ax
c     iprint : = 0 nothing is printed
c              = 1 the initial residual and the final relative residual
c                  are printed
c              = 2 the inial residual and all the relative residuals are
c                  printed
c                                       
c    RESULT:
c      
c    x : solution of Ax=b
c                                                
c    WORK ARRAYS :
c      
c    work  : it's a work array I need for gmres
c            the dimension of this array must be, at least of
c            idim*(k0+1)+(k0*(k0+1))/2+ 4*k0+2
c    yours : RHS of the equation (i.e., vector b)
c                                                   
c*************************************************************************
      implicit real*8 (a-h,o-z)
      integer   eamat                                       ! pointer
      integer   nevaw,nnode,nnodt,nnodw,ndofn,npoin,npois,nelem,lengp
      integer   idim, k0, itmax, iprint
      integer   lu, luj, luipl1, lsin, lcos, le, ly, lwork
      integer   lh, lhi, lhip1i, lhji, iter
      integer   ksymm,lures,nwork,lnods(nnodt,nelem),iffix(idim)
      real*8    resrel,rinit,reslast
      real*8    x(idim), work(nwork), yours(idim)
      real*8    diag(idim), elpve(nevaw), elwve(nevaw), work1(idim,k0+1)

      do i=1,idim
        if (abs(diag(i)).lt.1.d-15) diag(i)=1.0d0
      end do
c
c*** Memory allocation for the basis of Krylov subspaces, for the Hessenberg
c*** matrix, for cos(teta) and sin(teta), for y and e
c      
      lu   = 1
      lh   = 1
      lcos = lh+(k0*(k0+1))/2+1
      lsin = lcos+k0
      le   = lsin+k0
      ly   = le+k0+1
      lwork= ly+k0
      iter = 0
c
c*** Reference norm to compare the residuals (norm of the RHS excluding
c*** the prescribed degrees of freedom) and preconditioning     
c
      rinit=0.0d0
      do i = 1, idim
        yours(i) = yours(i)/diag(i)
        if(iffix(i).ne.1) rinit=rinit+yours(i)*yours(i)
      end do
      rinit=sqrt(rinit)
      residu=rinit
c
c*** Label to define the external loop
c
   20 continue
c      
c*** Compute the first residual from zero or from the last solution obtained
c      
      call gmrabx(eamat,lengp,x,diag,idim,work1(1,lu),yours, 
     .                1,lnods,elpve,elwve,nevaw,nnode,nnodt,
     .            nnodw,ndofn,npoin,npois,nelem,ksymm)

      prod=svecdo(work1(1,lu),work1(1,lu),idim)
      prod   = dsqrt(prod)
      prodin = 1.0d0/prod
      call svecbs(idim,work1(1,lu),work1(1,lu),prodin)
      work(le) = prod
      if ((iprint.gt.0).and.(iter.eq.0)) then
        write(lures,100) prod/rinit        
      endif
c
c*** Initialisation of the error vector
c      
      do i = 1, k0
        work(le+i) = 0.d0
      end do

      lhi = lh
      i = 0
c
c*** Label to define the internal loop : minimisation of the residual
c*** on a Krylov subspace
c      
   35 i = i+1
      luipl1 = lu+i
      call gmrabx(eamat,lengp,work1(1,luipl1-1),diag,idim,
     .            work1(1,luipl1),yours,0,
     .            lnods,elpve,elwve,nevaw,nnode,nnodt,nnodw,ndofn,
     .            npoin,npois,nelem,ksymm)
c      
c*** Gram-Schmidt orthogonalisation and construction of the column number
c*** i of the Hessenberg matrix :
c
      do j = 1, i
        luj  = lu+(j-1)
        lhji = lhi+j-1
        work(lhji)=svecdo(work1(1,luipl1),work1(1,luj),idim)
        coef = -work(lhji)
        call svecco(1.d0,work1(1,luipl1),coef,work1(1,luj),
     .              work1(1,luipl1),idim)
      end do

      lhip1i = lhi+i  
      work(lhip1i)=svecdo(work1(1,luipl1),work1(1,luipl1),idim)
      work(lhip1i) = dsqrt(work(lhip1i))
      
      coef = 1.d0/work(lhip1i)
      call svecbs(idim,work1(1,luipl1),work1(1,luipl1),coef)
c                                     
c*** QR factorisation of the Hessenberg matrix
c
      do j = 1, i-1
        lhji         = lhi+j-1
        lhjp1i       = lhi+j
        hji          = work(lhji)   
        hjp1i        = work(lhjp1i) 
        work(lhji)   = work(lcos+j-1)*hji+work(lsin+j-1)*hjp1i
        work(lhjp1i) = work(lcos+j-1)*hjp1i-work(lsin+j-1)*hji
      end do          
      hii            = work(lhi+i-1)
      hip1i          = work(lhi+i)  
      sqrhi          = dsqrt(hii*hii+hip1i*hip1i)
      work(lcos+i-1) = hii/sqrhi
      work(lsin+i-1) = hip1i/sqrhi
      work(lhi+i-1)  = sqrhi
      ei             = work(le+i-1)
      work(le+i-1)   = work(lcos+i-1)*ei
      work(le+i)     = -work(lsin+i-1)*ei
      reslast        = residu
      residu         = abs(work(le+i))/rinit
      iter = iter+1
      lhi = lhi+i
c
c*** End of the internal loop
c      
      if (((residu.gt.toler).and.(iter.lt.itmax)).and.(i.lt.k0)) goto 35
c           
c*** Resolution of the upper triangular system
c
      call gmrsol(work(lh),work(ly),work(le),i)
c                        
c*** Computation of the new estimation of the solution
c
      do j = 1, i
        luj= lu+(j-1)
        call svecco(1.d0,x,work(ly+j-1),work1(1,luj),x,idim)
      end do
c                              
c*** Test for reinitialization (if positive, repeat external loop)
c      
      if ((residu.gt.toler).and.(iter.lt.itmax)) goto 20
      resrel=residu/reslast
      if (iprint.gt.0) write(lures,110) residu,resrel,iter
      
  100 format(11x,'*** INITIAL RESIDUAL NORM  : ',e12.5)
  110 format(11x,'*** FINAL RESIDUAL NORM    : ',e12.5,/,
     .       11x,'*** LAST RESIDUAL FACTOR   : ',e12.5,/,
     .       11x,'*** NUMBER OF ITERATIONS   : ',i12)
      
      end
