      subroutine ite_ee(
     .  iffix,lnods,unkno,rhsid,nnode,nnodt,
     .  nnodw,nelem,ndofn,ntotv,npoin,nsist,
     .  iprob,kfact,outso)
c****************************************************************************
c
c**** This routine drives the library to solve a linear system
c**** of equations by EE iterative solvers 
c
c****************************************************************************
      implicit none
      include  'MatMan.h'
      integer   nnode,nnodt,nnodw,nelem,ndofn,npoin,ntotv,
     .          nsist,nevat,iprob,kfact,isist,itotv
      integer   iffix(ntotv),        lnods(nnodt,nelem)
      real*8    unkno(ntotv,nsist),  rhsid(ntotv,nsist)
      logical*1 outso
      integer   p_mat,p_rhs,p_for,p_pre,iwoso               ! pointer
      integer   lbya1,lbya2,lbyt1,lbyt2,svomt
      data      lbya1,lbya2,svomt/0,0,0/
      save      lbya1,lbya2,iwoso,p_for,p_pre
c
c*** Initializations
c
      nevat=ndofn*nnode+nnodw-nnode
c
c*** Store the original RHS and drive auxiliar memory
c
      lbyt1=8*ntotv*nsist
      if (lbya1.eq.0) then
        lbya1=lbyt1
        call sadmem(0,iwoso,lbya1,svomt)
      else
        if (lbyt1.gt.lbya1) then
          svome(iprob)=max(svome(iprob),svomt)
          call sadmem(2,iwoso,lbya1,svomt)
          lbya1=lbyt1
          call sadmem(0,iwoso,lbya1,svomt)
        end if
      end if

      lbyt2=8*nevat
      if (lbya2.eq.0) then
        lbya2=lbyt2
        call sadmem(0,p_for,lbya2,svomt)
        call sadmem(0,p_pre,lbya2,svomt)
      else
        if (lbyt2.gt.lbya2) then
          svome(iprob)=max(svome(iprob),svomt)
          call sadmem(2,p_pre,lbya2,svomt)
          call sadmem(2,p_for,lbya2,svomt)
          lbya2=lbyt2
          call sadmem(0,p_for,lbya2,svomt)
          call sadmem(0,p_pre,lbya2,svomt)
        end if
      end if
        
      call slveca(ntotv*nsist,rhsid,%val(iwoso))
c
c*** If needed, prescribe Dirichlet boundary conditions
c
      call spoias(%val(irhsx+(iprob-1)*lengp),p_rhs)
      call spoias(%val(iamat+(iprob-1)*lengp),p_mat)
      if(p_rhs.ne.0) then                                   ! kimpo was not 0
        if(kfact.eq.1) then                                 ! A x_p computed
          call iterhs(iffix,lnods,rhsid,%val(p_rhs),
     .                unkno,%val(p_pre),%val(p_for),
     .                ntotv,nsist,nnode,nnodt,nnodw,
     .                ndofn,nevat,nelem,npoin,lengp,
     .                p_mat,ksymm(iprob),zeroc) 
        end if
        do isist=1,nsist
          do itotv=1,ntotv
            if(iffix(itotv).eq.1) rhsid(itotv,isist)=0.0
          end do
        end do
        call svecad(nsist*ntotv,rhsid,%val(p_rhs),rhsid)
      end if
c
c*** Conjugate gradient solver
c
      if (kites(iprob).eq.21) then
        do isist=1,nsist
          call conjgr(lnods,p_mat,unkno(1,isist),rhsid(1,isist),
     .                toler(iprob),itmax(iprob),nnode,nnodt,nnodw,
     .                nelem,nequa(iprob),ndofn,npoin,lengp,iffix,
     .                lusol,svome(iprob),svomt,kmsip(iprob),outso)
        end do
c
c*** GMRES solver
c
      else if (kites(iprob).eq.22) then
        do isist=1,nsist
          call gmress(p_mat,nequa(iprob),rhsid(1,isist),unkno(1,isist),
     .                toler(iprob),itmax(iprob),kkryl(iprob),lnods,
     .                nnode,nnodt,nnodw,nelem,ndofn,npoin,ksymm(iprob),
     .                lengp,iffix,lusol,svome(iprob),svomt,kmsip(iprob),
     .                outso)
        end do
      end if
c
c*** Restore the original RHS and deallocate the work space used for it
c
      call slveca(nsist*ntotv,%val(iwoso),rhsid)
c
c*** Write temporary memory requirements 
c
      svome(iprob)=max(svome(iprob),svomt)
      if(outso) write(lusol,120) svome(iprob)
c
c*** Formats
c
  120 format(11x,'*** VOLATILE MEMORY (BYTES): ',i12)
      
      end
