      subroutine itdriv(n,ns,rhs,sol,ipar,fpar,a,ja,ia,
     .  sotyp,iou,svomt,lenin,lenre,outso)
c******************************************************************************
c                                                                             *
c**** This routine drives the iterative solver to be used. The input is:      *
c                                                                             *
c     n:          size of the linear system                                   *
c     ns:         number of systems to be solved. Maximum = 10.               *
c     rhs(n,ns):  right-hand side of the linear system                        *
c     sol(n,ns):  solution to the linear system                               *
c     ipar(16):   integer parameter array for the reverse communication       *
c                 protocol. The components supplied on input are:             *
c                 ipar(2) < 0 No preconditioning                              *
c                         = 0 Diagonal preconditioning                        *
c                         = 1 Left preconditioning only                       *
c                         = 2 Right preconditioning only                      *
c                         = 3 Both left and right preconditioning             *
c                 ipar(5) = Size the Krylov subsapce for FOM, GMRES,          *
c                           FGMRES and DQGMRES                                *
c                 ipar(6) = maximum number of matrix-vector multiplies,       * 
c                           if not a positive number the iterative solver     *
c                           will run till convergence test is satisfied.      *
c                 ipar(13)= The fill-in parameter for the ILU precondi-       *
c                           tioner. Each row of L and each row of U           *
c                           will have a maximum of ipar(13) elements in       *
c                           addition to their original number of nonzero      *
c                           elements.                                         *
c     fpar(16):   floating-point parameter array storing information to       *
c                 and from the iterative solvers. The components supplied     *
c                 on input are:                                               *
c                 fpar(1) = relative convergence tolerance                    *
c                 fpar(13)= Thresholding tolerance in L and U. Any element    *
c                           whose size is less than this tolerance (relative  *
c                           to the norm of current row in U) is dropped (=0   *
c                           is allowed).                                      *
c     a           Matrix A of the system to be solved stored in CSR format    *
c     ja          Column indices of A                                         *
c     ia          Pointers to the beginning of each row of A                  *
c     sotyp       Integer with the solver type to be used (see above)         *
c     iou         Output unit                                                 *
c     svomt       Counter of volatile memory employed by the code (maximum)   *
c     lenin       Length of integers and pointers                             *
c     lenre       Length of reals                                             *
c     outso       Logical flag to decide whether to write the output of not   *
c                                                                             *
c******************************************************************************
      implicit none
      integer
     .  n,ns,sotyp,iou,svomt,oldme,newme,lenin,lenre,
     .  ipar(16), ia(n+1), ja(*)
      real*8
     .  fpar(16),rhs(n,ns),sol(n,ns),a(*)
      integer
     .  is, its, nwk, ierr, lbyt1, lbyt2, lbyt3,
     .  lbyt4,lbyt5,l1,l2,lby1o,lby2o,lby3o,lby4o,lby5o
      integer
     .  p_wk, p_au, p_jau, p_ju, p_di                       ! Pointers
      logical*1
     .  again,outso,tosol,dosys(10)
      real*8
     .  res,resini,rhsno(10),vecdo
      external
     .  vecdo
      data
     .  oldme,lby1o,lby2o,lby3o,lby4o,lby5o/0,0,0,0,0,0/
      save
     .  oldme,lby1o,lby2o,lby3o,lby4o,lby5o,its,res,
     .  p_wk,p_au, p_jau, p_ju, p_di
c
c***  Initializations:
c***  Convergence check: | residual | < rtol * | rhs | + atol
c***  where atol = fpar(2)
c
      newme=0
      ipar(3) = 2                                           ! Type of conv. test 
      fpar(2) = 1.0d-5*fpar(1)                              ! ATOL for conv.
      if(ns.gt.10) call srunen(
     .  'ITDRIV: TOO MANY SYSTEMS',iou)
c
c***  Memory allocation
c
      if((sotyp.eq.1).or.(sotyp.eq.2)) then
        ipar(4) = 5*n
      else if(sotyp.eq.3) then
        ipar(4) = 7*n
      else if((sotyp.eq.4).or.(sotyp.eq.6)) then
        ipar(4) = 11*n
      else if(sotyp.eq.5) then
        ipar(4) = 8*n
      else if((sotyp.eq.7).or.(sotyp.eq.8)) then
        ipar(4) = (n+3)*(ipar(5)+2) + (ipar(5)+1)*ipar(5)/2
      else if(sotyp.eq.9) then
        ipar(4) = n*(2*ipar(5)+1) + (ipar(5)+1)*ipar(5)/2
     .    + 3*ipar(5) + 2 
      else if(sotyp.eq.10) then
        ipar(4) =n + (ipar(5)+1)*(2*n+4)
      end if
      lbyt1=lenre*ipar(4)

      call maxmem(p_wk,lbyt1,lby1o,newme)
c
c***  Initial residual (for the scaled matrix, if needed)
c
      if(ipar(2).ge.0) then
        lbyt5=n*lenre
        call maxmem(p_di,lbyt5,lby5o,newme)
        call diapre(%val(p_di),sol,a,ia,ja,rhs,n,iou,1)     !A' = D^-1/2 A D^-1/2 
      end if
      tosol = .false.
      do is=1,ns
        resini = 0.0
        rhsno(is) = sqrt(vecdo(n,rhs(1,is),rhs(1,is)))
        if(rhsno(is).gt.1.0e-10) then
          call amux  (
     .      n,sol(1,is),%val(p_wk),a,ja,ia)                 ! Ax
          call sveccl(
     .      n,1.0,%val(p_wk),-1.0,rhs(1,is),%val(p_wk))     ! Ax - f
          resini=sqrt(
     .      vecdo(n,%val(p_wk),%val(p_wk)))/rhsno(is)
        end if
        if(resini.le.fpar(1)) then
          dosys(is)=.false.
          if(outso) write (iou,100) resini
        else
          dosys(is)=.true.
          tosol=.true.
        end if
      end do
      if(.not.tosol) then
        if(ipar(2).ge.0) then
          do is = 1,ns
            call diapre(
     .        %val(p_di),sol(1,is),a,ia,ja,rhs,n,iou,2)
          end do
        end if
        return
      end if
c
c***  Set-up the preconditioner (after a diagonal scaling)
c
      if(ipar(2).gt.0) then
        lbyt2 =  (n+1)*lenin
        call maxmem(p_ju,lbyt2,lby2o,newme)
        nwk = n*ipar(13) - n                                ! 1st estimate ( - n)
        ierr = -3
        do while((ierr.eq.-2).or.(ierr.eq.-3)) 
          nwk = nwk + n
          lbyt3 = lenre*nwk
          lbyt4 = lenin*nwk
          call maxmem(p_au ,lbyt3,lby3o,newme)
          call maxmem(p_jau,lbyt4,lby4o,newme)
          l1 = lenre*n
          l2 = lenin*n
          call ilut(n, a, ja, ia, ipar(13), fpar(13),
     .      %val(p_au), %val(p_jau), %val(p_ju), nwk,
     .      %val(p_wk), %val(p_wk +l1),
     .      %val(p_wk + 2*l1         ), 
     .      %val(p_wk + 2*l1 +  l2), 
     .      %val(p_wk + 2*l1 +2*l2), 
     .      ierr)
        end do
        if(ierr.ne.0)
     .    call srunei(
     .    'ITDRIV: ERROR IN THE ILU PREC., CODE = ',
     .    ierr,iou)
      end if

      do is = 1,ns
        if(dosys(is)) then
c
c***  Normal execution
c
          its = 0
          res = 0.0d0
          ipar(1) = 0
          if(sotyp.eq.1) then                               !CG method
            again=.true.
            do while(again)
              call cg    (
     .          n,rhs(1,is),sol(1,is),ipar,fpar,%val(p_wk))
              call auxdri(
     .          n,ipar,fpar,its,res,iou,a,ja,ia,
     .          %val(p_au),%val(p_jau),%val(p_ju),
     .          %val(p_wk),again)
            end do
          else if(sotyp.eq.2) then                          !CGNR method
            again=.true.
            do while(again)
              call cgnr  (
     .          n,rhs(1,is),sol(1,is),ipar,fpar,%val(p_wk))
              call auxdri(
     .          n,ipar,fpar,its,res,iou,a,ja,ia,
     .          %val(p_au),%val(p_jau),%val(p_ju),
     .          %val(p_wk),again)
            end do
          else if(sotyp.eq.3) then                          !BCG method
            again=.true.
            do while(again)
              call bcg   (
     .          n,rhs(1,is),sol(1,is),ipar,fpar,%val(p_wk))
              call auxdri(
     .          n,ipar,fpar,its,res,iou,a,ja,ia,
     .          %val(p_au),%val(p_jau),%val(p_ju),
     .          %val(p_wk),again)
            end do
          else if(sotyp.eq.4) then                          !DBCG method
            again=.true.
            do while(again)
              call dbcg  (
     .          n,rhs(1,is),sol(1,is),ipar,fpar,%val(p_wk))
              call auxdri(
     .          n,ipar,fpar,its,res,iou,a,ja,ia,
     .          %val(p_au),%val(p_jau),%val(p_ju),
     .          %val(p_wk),again)
            end do
          else if(sotyp.eq.5) then                          !BCGSTAB method
            again=.true.
            do while(again)
              call bcgstab(
     .          n,rhs(1,is),sol(1,is),ipar,fpar,%val(p_wk))
              call auxdri (
     .          n,ipar,fpar,its,res,iou,a,ja,ia,
     .          %val(p_au),%val(p_jau),%val(p_ju),
     .          %val(p_wk),again)
            end do
          else if(sotyp.eq.6) then                          !TFQMR method
            again=.true.
            do while(again)
              call tfqmr (
     .          n,rhs(1,is),sol(1,is),ipar,fpar,%val(p_wk))
              call auxdri(
     .          n,ipar,fpar,its,res,iou,a,ja,ia,
     .          %val(p_au),%val(p_jau),%val(p_ju),
     .          %val(p_wk),again)
            end do
          else if(sotyp.eq.7) then                          !FOM method
            again=.true.
            do while(again)
              call fom   (
     .          n,rhs(1,is),sol(1,is),ipar,fpar,%val(p_wk))
              call auxdri(
     .          n,ipar,fpar,its,res,iou,a,ja,ia,
     .          %val(p_au),%val(p_jau),%val(p_ju),
     .          %val(p_wk),again)
            end do
          else if(sotyp.eq.8) then                          !GMRES method
            again=.true.
            do while(again)
              call gmres (
     .          n,rhs(1,is),sol(1,is),ipar,fpar,%val(p_wk))
              call auxdri(
     .          n,ipar,fpar,its,res,iou,a,ja,ia,
     .          %val(p_au),%val(p_jau),%val(p_ju),
     .          %val(p_wk),again)
            end do
          else if(sotyp.eq.9) then                          !FGMRES method
            again=.true.
            do while(again)
              call fgmres(
     .          n,rhs(1,is),sol(1,is),ipar,fpar,%val(p_wk))
              call auxdri(
     .          n,ipar,fpar,its,res,iou,a,ja,ia,
     .          %val(p_au),%val(p_jau),%val(p_ju),
     .          %val(p_wk),again)
            end do
          else if(sotyp.eq.10) then                         !DQGMRES method
            again=.true.
            do while(again)
              call dqgmres(
     .          n,rhs(1,is),sol(1,is),ipar,fpar,%val(p_wk))
              call auxdri (
     .          n,ipar,fpar,its,res,iou,a,ja,ia,
     .          %val(p_au),%val(p_jau),%val(p_ju),
     .          %val(p_wk),again)
            end do
          end if
c
c*** Write to output file and recover of the unknown
c
          if(ipar(2).ge.0)                                  !x = D^-1/2 x' 
     .      call diapre(
     .      %val(p_di),sol(1,is),a,ia,ja,rhs,n,iou,2)
          if(outso) write (iou,110) fpar(3)/rhsno(is),
     .      fpar(6)/rhsno(is),fpar(7),ipar(7)
        else
          if(ipar(2).ge.0)
     .      call diapre(
     .      %val(p_di),sol(1,is),a,ia,ja,rhs,n,iou,2)
        end if
      end do

  100 format(11x,'*** CONVERGED GUESS, ERROR : ',e12.5)
  110 format(11x,'*** INITIAL RESIDUAL NORM  : ',e12.5,/,
     .       11x,'*** FINAL RESIDUAL NORM    : ',e12.5,/,
     .       11x,'*** CONVERGENCE RATE       : ',e12.5,/,
     .       11x,'*** NUMBER OF ITERATIONS   : ',i12)
c
c***  To deallocate memory, use the lines commented out
c
      if(oldme.lt.newme) then
        oldme = newme
        svomt = svomt + newme - oldme
      end if

c-m   call sadmem(2,p_wk ,lbyt1,svomt)
c-m   if(ipar(2).gt.0) then
c-m     call sadmem(2,p_ju ,lbyt2,svomt)
c-m     call sadmem(2,p_au ,lbyt3,svomt)
c-m     call sadmem(2,p_jau,lbyt4,svomt)
c-m     call sadmem(2,p_di ,lbyt5,svomt)
c-m   end if

      end
