      subroutine renumb(
     .  lnods,iffix,nnode,nnodt,npoin,nelem,ndofn,
     .  ntotv,iprob)
c****************************************************************************
c
c**** This routine defines the address of the array LPNTN (ialpn) for
c**** the problem we are considering and sets up the components of this
c**** array according to the renumbering strategy used. The basic tree is
c
c     RENUMB                Main routine
c           CONDRE          Modification of LNODS (for condensed d.o.f.)
c           RENBAN          Renumbering for band storage
c                 SETCON    Set connectivities
c                 OPTBAN    Optimize bandwidth
c                 CONDCO    Undo the original modification for condensation
c           RENUMN          Renumbering for profile storage
c                 RENUM0    Optimize profile
c                 CONDCO    Undo the original modification for condensation
c           RENBAN          Renumbering for band storage (used for sparse,
c                           only to minimize the fill-in)      
c           SKYINI          Construct LPONT (profile storage)
c
c****************************************************************************
      implicit none
      include   'MatMan.h'
      integer    iwoso,p_lpn,p_lpo,p_lnw,p_tou              ! pointer
      integer    nnode,nnodt,npoin,nelem,ndofn,ntotv,iprob,
     .           npois,nnodw,nevat,npoiw,npows,ldumy,lbyts,
     .           inode,ielem,npoit,svomt,ipass
      integer    lnods(nnodt,nelem), iffix(ntotv)
      data       svomt,ipass/0,0/
      save       svomt,ipass
c
c*** Check parameters, allocate memory for pointers and initialize flags
c
      if(ipass.eq.0) then
        call almepo
        call setdat
        call slvecz(mprob,svome) 
        call slvecz(mprob,speme)
        call slvecz(mprob,sizms)
        ipass = 1
      end if
c
c*** Calculation of some integers
c
      npois=ntotv-ndofn*npoin                     ! Points with 1 d.o.f.
      npoiw=npoin                                 ! Main working points
      npows=npois                                 ! Secondary working points
      nnodw=nnode                                 ! Working nodes
      nevat=ndofn*nnode                           ! Working d.o.f.
      npoit=npoiw+npows                           ! Total number of points
      if (npois.gt.0) then
        nnodw=nnodt
        nevat=nevat+nnodt-nnode
      end if
c
c*** Memory allocation for LPNTN and LPONT
c
      if ((sofam(iprob).eq.1).or
     .  .(sofam(iprob).eq.3)) then
        lbyts=lengp*ntotv
        call sadmem(0,p_lpn,lbyts,speme(iprob))
        call sadmem(0,p_lpo,lbyts,speme(iprob))
        call slvecz(ntotv,%val(p_lpn))
        call slvecz(ntotv,%val(p_lpo))
        call spoias(
     .    p_lpn,%val(ialpn+(iprob-1)*lengp))
        call spoias(
     .    p_lpo,%val(ialpo+(iprob-1)*lengp))
      end if
c
c*** Modification of LNODS if some degrees of freedom have been
c*** condensed
c      
      if (conde(iprob).and.
     .  (krenu(iprob).gt.0)) then
        lbyts=lengp*nelem*nnodt
        call sadmem(0,p_lnw,lbyts,svomt)                    ! Memory for LNODW
        call condre(
     .    nnode,nnodt,nnodw,nelem,npoit,                    ! Use LPONT as 
     .    npoiw,npows,lnods,%val(p_lnw),                    ! auxiliar array
     .    %val(p_lpo))
      end if
c
c*** Renumber strategy according to the type of matrix storage for direct
c*** solvers (SOFAM = 1). For sk iterative solvers (SOFAM = 3) the routine
c*** RENBAN is used to initialize LPNTN, which is employed for identifying
c*** prescribed and free degrees-of-freedom      
c
      if(sofam(iprob).eq.1) then
        if (kmsip(iprob).eq.0) then                         ! BAND STORAGE
          call renban(
     .      lnods,%val(p_lpn),%val(p_lpo),
     .      iffix,nnodw,nnodt,npoin,npoiw,npows,
     .      nelem,ndofn,npois,ntotv,nequa(iprob),
     .      lusol,krenu(iprob),lengp,conde(iprob),
     .      svome(iprob),svomt)
        else if (kmsip(iprob).eq.1) then                    ! PROFILE
          call renumn(
     .      lnods,%val(p_lpn),%val(p_lpo), 
     .      iffix,nnodw,nnodt,npoin,npoiw,npows,
     .      nelem,ndofn,npois,ntotv,nequa(iprob),
     .      lusol,krenu(iprob),lengp,conde(iprob),
     .      svome(iprob),svomt)
        else if (kmsip(iprob).eq.2) then                    ! SPARSE STORAGE
          call renban(
     .      lnods,%val(p_lpn),%val(p_lpo),
     .      iffix,nnodw,nnodt,npoin,npoiw,npows,
     .      nelem,ndofn,npois,ntotv,nequa(iprob),
     .      lusol,krenu(iprob),lengp,conde(iprob),
     .      svome(iprob),svomt)
        end if
      else if(sofam(iprob).eq.2) then
        nequa(iprob)=ntotv
      else if(sofam(iprob).eq.3) then
        call renban(
     .    lnods,%val(p_lpn),%val(p_lpo),
     .    iffix,nnodw,nnodt,npoin,npoiw,npows,
     .    nelem,ndofn,npois,ntotv,nequa(iprob),
     .    lusol,    0,lengp,conde(iprob),
     .    svome(iprob),svomt)
      end if
c
c*** Recover LNODS and deallocate memory for LNODW, if needed
c
      if (conde(iprob).and.(krenu(iprob).gt.0)) then
        call slveca(nnodt*nelem,%val(p_lnw),lnods)
        svome(iprob) = max(svome(iprob),svomt)
        call sadmem(2,p_lnw,lengp*nnodt*nelem,svomt)
      end if
c
c*** Construct LPONT in the case of a slyline matrix storage
c      
      if ((sofam(iprob).eq.1).and.(kmsip(iprob).eq.1)) then
        call sadmem(0,iwoso,lengp*nevat,ldumy)
        call skyini(iprob,lnods,ndofn,nelem,        
     .              nequa(iprob),nevat,nnodt,nnodw, 
     .              npoin,ntotv,%val(p_lpn),
     .              %val(p_lpo),%val(iwoso))
        call sadmem(2,iwoso,lengp*nevat,ldumy)
      end if
c
c*** This routine can be called to check the renumber strategy
c
c-k   if(krenu(iprob).gt.0) then
c-k     call sadmem(0,p_tou,ntotv,svomt)
c-k     call cheren(%val(p_lpn),%val(p_tou),lusol,ntotv)
c-k     svome(iprob) = max(svome(iprob),svomt)
c-k     call sadmem(2,p_tou,ntotv,svomt)
c-k   end if
c
c*** Put all values of LNODS positive, if needed
c
      if(conde(iprob)) then
        do ielem=1,nelem
          do inode=1,nnodt
            lnods(inode,ielem)=abs(lnods(inode,ielem))
          end do
        end do
      end if

      end
