      subroutine skyini(
     .  iprob,lnods,ndofn,nelem,neqns,nevac,
     .  nnodt,nnodw,npoin,ntotv,lnueq,lpont,
     .  leqns)
c***********************************************************************
c
c**** This routine performs the skyline solver initialization when
c**** renumbering is employed
c
c.... Input parameters
c
c     lnods(nnodt,nelem)  -  element connectivity
c     lnueq(ntotv)        -  equation numbers
c     ndofn               -  no. of d.o.f per node
c     nelem               -  no. of elements
c     neqns               -  number of equations
c     nnodt               -  number of element nodes
c     npoin               -  number of points
c     ntotv               -  number of variables
c     nnodw               -  number of nodes to work
c
c.... Output parameters
c
c     lpont(neqns)        -  position of the term just above the diagonal
c                            in the arrays gstup & gstlo
c
c.... Auxiliary working parameters
c
c     leqns(nevac)        -  auxiliary array to store the equation
c                            numbers of the current element
c
c***********************************************************************
      implicit real*8(a-h,o-z)
      include 'MatMan.h'
      integer
     .  lpont(neqns), leqns(nevac),
     .  lnueq(ntotv), lnods(nnodt,nelem)
c
c*** Find the position of the diagonal terms
c
      do ieqns=1,neqns
        lpont(ieqns)=0
      end do

      maxpr = 0
      do ielem = 1,nelem
        mneqn = 0
        nevac = 0
        do inode = 1,nnodw
          ipont = lnods(inode,ielem)
          if (ipont.gt.0) then
            if (ipont.le.npoin) then
              itotv=(ipont-1)*ndofn
              do idofn = 1,ndofn
                itotv=itotv+1
                ieqns = lnueq(itotv)
                if(ieqns.gt.0) then
                  if(mneqn.eq.0) mneqn = ieqns
                  mneqn = min0(mneqn,ieqns)
                  nevac = nevac + 1
                  leqns(nevac) = ieqns
                end if
              end do
            else if (ipont.ne.0) then
              itotv=npoin*ndofn+(ipont-npoin)
              ieqns = lnueq(itotv)
              if(ieqns.gt.0) then
                if(mneqn.eq.0) mneqn = ieqns
                mneqn = min0(mneqn,ieqns)
                nevac = nevac + 1
                leqns(nevac) = ieqns
              end if
            end if
          end if
        end do

        if(nevac.gt.0) then
          do ievac = 1,nevac
            ieqns = leqns(ievac)
            npont = max0(lpont(ieqns),ieqns-mneqn)
            lpont(ieqns) = npont
            maxpr = max0(maxpr,lpont(ieqns))
          end do
        end if
      end do ! ielem = 1,nelem
c
c*** Compute diagonal pointers for profile
c
      lpont(1) = 0
      if(neqns.gt.1) then
        do ieqns = 2,neqns
          lpont(ieqns) = lpont(ieqns) + lpont(ieqns-1)
        end do
      end if

      end
